import os
from typing import Optional

from feldera.rest.feldera_client import FelderaClient
from feldera.rest.pipeline import Pipeline as InnerPipeline
from feldera.pipeline import Pipeline
from feldera.enums import CompilationProfile, PipelineFieldSelector
from feldera.runtime_config import RuntimeConfig
from feldera.rest.errors import FelderaAPIError


class PipelineBuilder:
    """
    A builder for creating a Feldera Pipeline.

    :param client: The :class:`.FelderaClient` instance
    :param name: The name of the pipeline
    :param description: The description of the pipeline
    :param sql: The SQL code of the pipeline
    :param udf_rust: Rust code for UDFs
    :param udf_toml: Rust dependencies required by UDFs (in the TOML format)
    :param compilation_profile: The :class:`.CompilationProfile` to use
    :param runtime_config: The :class:`.RuntimeConfig` to use. Enables
        configuring the runtime behavior of the pipeline such as:
        fault tolerance, storage and :class:`.Resources`
    """

    def __init__(
        self,
        client: FelderaClient,
        name: str,
        sql: str,
        udf_rust: str = "",
        udf_toml: str = "",
        description: str = "",
        compilation_profile: CompilationProfile = CompilationProfile.OPTIMIZED,
        runtime_config: RuntimeConfig = RuntimeConfig.default(),
        runtime_version: Optional[str] = None,
    ):
        self.client: FelderaClient = client
        self.name: str | None = name
        self.description: str = description
        self.sql: str = sql
        self.udf_rust: str = udf_rust
        self.udf_toml: str = udf_toml
        self.compilation_profile: CompilationProfile = compilation_profile
        self.runtime_config: RuntimeConfig = runtime_config
        self.runtime_version: Optional[str] = os.environ.get(
            "FELDERA_RUNTIME_VERSION", runtime_version
        )

    def create(self) -> Pipeline:
        """
        Create the pipeline if it does not exist.

        :return: The created pipeline
        """

        if self.name is None or self.sql is None:
            raise ValueError("Name and SQL are required to create a pipeline")

        try:
            if (
                self.client.get_pipeline(self.name, PipelineFieldSelector.STATUS)
                is not None
            ):
                raise RuntimeError(f"Pipeline with name {self.name} already exists")
        except FelderaAPIError as err:
            if err.error_code != "UnknownPipelineName":
                raise err

        inner = InnerPipeline(
            self.name,
            description=self.description,
            sql=self.sql,
            udf_rust=self.udf_rust,
            udf_toml=self.udf_toml,
            program_config={
                "profile": self.compilation_profile.value,
                "runtime_version": self.runtime_version,
            },
            runtime_config=self.runtime_config.to_dict(),
        )

        inner = self.client.create_pipeline(inner)
        pipeline = Pipeline(self.client)
        pipeline._inner = inner

        return pipeline

    def create_or_replace(self) -> Pipeline:
        """
        Creates a pipeline if it does not exist and replaces it if it exists.

        If the pipeline exists and is running, it will be stopped and replaced.
        """

        if self.name is None or self.sql is None:
            raise ValueError("Name and SQL are required to create a pipeline")

        try:
            # shutdown the pipeline if it exists and is running
            p = Pipeline.get(self.name, self.client)
            p.stop(force=True)
            p.clear_storage()

        except FelderaAPIError:
            # pipeline doesn't exist, no worries
            pass

        inner = InnerPipeline(
            self.name,
            description=self.description,
            sql=self.sql,
            udf_rust=self.udf_rust,
            udf_toml=self.udf_toml,
            program_config={
                "profile": self.compilation_profile.value,
                "runtime_version": self.runtime_version,
            },
            runtime_config=self.runtime_config.to_dict(),
        )

        inner = self.client.create_or_update_pipeline(inner)
        pipeline = Pipeline(self.client)
        pipeline._inner = inner

        return pipeline



import uuid

import pandas as pd
from decimal import Decimal


def sql_type_to_pandas_type(sql_type: str):
    """
    Converts a SQL type to a pandas type.
    """

    match sql_type.upper():
        case "UUID":
            return None
        case "BOOLEAN":
            return "boolean"
        case "TINYINT":
            return "Int8"
        case "SMALLINT":
            return "Int16"
        case "INTEGER":
            return "Int32"
        case "BIGINT":
            return "Int64"
        case "REAL":
            return "Float32"
        case "DOUBLE":
            return "Float64"
        case "DECIMAL":
            return None
        case "CHAR":
            return "str"
        case "VARCHAR":
            return "str"
        case "DATE" | "TIMESTAMP":
            return "datetime64[ns]"
        case "TIME" | "INTERVAL":
            return "timedelta64[ns]"
        case "ARRAY":
            return None
        case "NULL":
            return None
        case "BINARY" | "VARBINARY":
            return None
        case "STRUCT" | "MAP":
            return None


def ensure_dataframe_has_columns(df: pd.DataFrame):
    """
    Ensures that the DataFrame has column names set.
    """

    if [v for v in range(df.shape[1])] == list(df.columns):
        raise ValueError(
            """
            DataFrame has no column names set.
            Input DataFrame must have column names set and they must be consistent with the columns in the input table.
            """
        )


def dataframe_from_response(buffer: list[list[dict]], fields: list[dict]):
    """
    Converts the response from Feldera to a pandas DataFrame.
    """

    pd_schema = {}

    decimal_col = []
    uuid_col = []

    for column in fields:
        column_name = column["name"]
        if not column["case_sensitive"]:
            column_name = column_name.lower()
        column_type = column["columntype"]["type"]
        if column_type == "DECIMAL":
            decimal_col.append(column_name)
        elif column_type == "UUID":
            uuid_col.append(column_name)

        pd_schema[column_name] = sql_type_to_pandas_type(column_type)

    data = [
        {**item["insert"], "insert_delete": 1}
        if "insert" in item
        else {**item["delete"], "insert_delete": -1}
        for sublist in buffer
        for item in sublist
    ]

    if len(decimal_col) != 0:
        for datum in data:
            for col in decimal_col:
                if datum[col] is not None:
                    datum[col] = Decimal(datum[col])

    if len(uuid_col) != 0:
        for datum in data:
            for col in uuid_col:
                if datum[col] is not None:
                    datum[col] = uuid.UUID(datum[col])

    df = pd.DataFrame(data)
    df = df.astype(pd_schema)

    return df


def chunk_dataframe(df, chunk_size=1000):
    """
    Yield successive n-sized chunks from the given dataframe.
    """

    for i in range(0, len(df), chunk_size):
        yield df.iloc[i : i + chunk_size]



from threading import Thread
from typing import Callable, Optional

import pandas as pd
from feldera import FelderaClient
from feldera._helpers import dataframe_from_response
from feldera.enums import PipelineFieldSelector


class CallbackRunner(Thread):
    def __init__(
        self,
        client: FelderaClient,
        pipeline_name: str,
        view_name: str,
        callback: Callable[[pd.DataFrame, int], None],
        exception_callback: Callable[[BaseException], None],
    ):
        super().__init__()
        self.daemon = True
        self.client: FelderaClient = client
        self.pipeline_name: str = pipeline_name
        self.view_name: str = view_name
        self.callback: Callable[[pd.DataFrame, int], None] = callback
        self.exception_callback: Callable[[BaseException], None] = exception_callback
        self.schema: Optional[dict] = None

    def run(self):
        """
        The main loop of the thread. Listens for data and calls the callback function on each chunk of data received.

        :meta private:
        """

        try:
            pipeline = self.client.get_pipeline(
                self.pipeline_name, PipelineFieldSelector.ALL
            )

            schemas = pipeline.tables + pipeline.views
            for schema in schemas:
                if schema.name == self.view_name:
                    self.schema = schema
                    break

            if self.schema is None:
                raise ValueError(
                    f"Table or View {self.view_name} not found in the pipeline schema."
                )

            gen_obj = self.client.listen_to_pipeline(
                self.pipeline_name,
                self.view_name,
                format="json",
                case_sensitive=self.schema.case_sensitive,
            )

            iterator = gen_obj()

            for chunk in iterator:
                chunk: dict = chunk
                data: Optional[list[dict]] = chunk.get("json_data")
                seq_no: Optional[int] = chunk.get("sequence_number")
                if data is not None and seq_no is not None:
                    self.callback(
                        dataframe_from_response([data], self.schema.fields), seq_no
                    )
        except BaseException as e:
            self.exception_callback(e)



from enum import Enum
from typing import Optional


class CompilationProfile(Enum):
    """
    The compilation profile to use when compiling the program.
    """

    SERVER_DEFAULT = None
    """
    The compiler server default compilation profile.
    """

    DEV = "dev"
    """
    The development compilation profile.
    """

    UNOPTIMIZED = "unoptimized"
    """
    The unoptimized compilation profile.
    """

    OPTIMIZED = "optimized"
    """
    The optimized compilation profile, the default for this API.
    """


class BuildMode(Enum):
    CREATE = 1
    GET = 2
    GET_OR_CREATE = 3


class DeploymentDesiredStatus(Enum):
    """
    Deployment desired status of the pipeline.
    """

    STOPPED = 0
    UNAVAILABLE = 1
    STANDBY = 2
    PAUSED = 3
    RUNNING = 4
    SUSPENDED = 5

    @staticmethod
    def from_str(value):
        for member in DeploymentDesiredStatus:
            if member.name.lower() == value.lower():
                return member
        raise ValueError(
            f"Unknown value '{value}' for enum {DeploymentDesiredStatus.__name__}"
        )


class DeploymentResourcesDesiredStatus(Enum):
    """
    The desired status of deployment resources of the pipeline.
    """

    STOPPED = 0
    PROVISIONED = 1

    @staticmethod
    def from_str(value):
        for member in DeploymentResourcesDesiredStatus:
            if member.name.lower() == value.lower():
                return member
        raise ValueError(
            f"Unknown value '{value}' for enum {DeploymentResourcesDesiredStatus.__name__}"
        )


class DeploymentResourcesStatus(Enum):
    """
    The desired status of deployment resources of the pipeline.
    """

    STOPPED = 0
    PROVISIONING = 1
    PROVISIONED = 2
    STOPPING = 3

    @staticmethod
    def from_str(value):
        for member in DeploymentResourcesStatus:
            if member.name.lower() == value.lower():
                return member
        raise ValueError(
            f"Unknown value '{value}' for enum {DeploymentResourcesStatus.__name__}"
        )


class DeploymentRuntimeDesiredStatus(Enum):
    """
    Deployment runtime desired status of the pipeline.
    """

    UNAVAILABLE = 0
    STANDBY = 1
    PAUSED = 2
    RUNNING = 3
    SUSPENDED = 4

    @staticmethod
    def from_str(value):
        for member in DeploymentRuntimeDesiredStatus:
            if member.name.lower() == value.lower():
                return member
        raise ValueError(
            f"Unknown value '{value}' for enum {DeploymentRuntimeDesiredStatus.__name__}"
        )


class DeploymentRuntimeStatus(Enum):
    """
    Deployment runtime status of the pipeline.
    """

    UNAVAILABLE = 0
    STANDBY = 1
    AWAITINGAPPROVAL = 2
    INITIALIZING = 3
    BOOTSTRAPPING = 4
    REPLAYING = 5
    PAUSED = 6
    RUNNING = 7
    SUSPENDED = 8

    @staticmethod
    def from_str(value):
        for member in DeploymentRuntimeStatus:
            if member.name.lower() == value.lower():
                return member
        raise ValueError(
            f"Unknown value '{value}' for enum {DeploymentRuntimeStatus.__name__}"
        )


class PipelineStatus(Enum):
    """
    Represents the state that this pipeline is currently in.
    """

    NOT_FOUND = 0
    STOPPED = 1
    PROVISIONING = 2
    UNAVAILABLE = 3
    STANDBY = 4
    AWAITINGAPPROVAL = 5
    INITIALIZING = 6
    BOOTSTRAPPING = 7
    REPLAYING = 8
    PAUSED = 9
    RUNNING = 10
    SUSPENDED = 11
    STOPPING = 12

    @staticmethod
    def from_str(value):
        for member in PipelineStatus:
            if member.name.lower() == value.lower():
                return member
        raise ValueError(f"Unknown value '{value}' for enum {PipelineStatus.__name__}")

    def __eq__(self, other):
        return self.value == other.value


class TransactionStatus(Enum):
    """
    Represents the transaction handling status of a pipeline.
    """

    NoTransaction = 1
    """There is currently no active transaction."""

    TransactionInProgress = 2
    """There is an active transaction in progress."""

    CommitInProgress = 3
    """A commit is currently in progress."""

    @staticmethod
    def from_str(value):
        for member in TransactionStatus:
            if member.name.lower() == value.lower():
                return member
        raise ValueError(
            f"Unknown value '{value}' for enum {TransactionStatus.__name__}"
        )

    def __eq__(self, other):
        return self.value == other.value


class ProgramStatus(Enum):
    Pending = 1
    CompilingSql = 2
    SqlCompiled = 3
    CompilingRust = 4
    Success = 5
    SqlError = 6
    RustError = 7
    SystemError = 8

    def __init__(self, value):
        self.error: Optional[dict] = None
        self._value_ = value

    @staticmethod
    def from_value(value):
        error = None
        if isinstance(value, dict):
            error = value
            value = list(value.keys())[0]

        for member in ProgramStatus:
            if member.name.lower() == value.lower():
                member.error = error
                return member
        raise ValueError(f"Unknown value '{value}' for enum {ProgramStatus.__name__}")

    def __eq__(self, other):
        return self.value == other.value

    def __str__(self):
        return self.name + (f": ({self.error})" if self.error else "")

    def get_error(self) -> Optional[dict]:
        """
        Returns the compilation error, if any.
        """

        return self.error


class CheckpointStatus(Enum):
    Success = 1
    Failure = 2
    InProgress = 3
    Unknown = 4

    def __init__(self, value):
        self.error: Optional[str] = None
        self._value_ = value

    def __eq__(self, other):
        return self.value == other.value

    def get_error(self) -> Optional[str]:
        """
        Returns the error, if any.
        """

        return self.error


class StorageStatus(Enum):
    """
    Represents the current storage usage status of the pipeline.
    """

    CLEARED = 0
    """
    The pipeline has not been started before, or the user has cleared storage.

    In this state, the pipeline has no storage resources bound to it.
    """

    INUSE = 1
    """
    The pipeline was (attempted to be) started before, transitioning from `STOPPED`
    to `PROVISIONING`, which caused the storage status to become `INUSE`.

    Being in the `INUSE` state restricts certain edits while the pipeline is `STOPPED`.

    The pipeline remains in this state until the user invokes `/clear`, transitioning
    it to `CLEARING`.
    """

    CLEARING = 2
    """
    The pipeline is in the process of becoming unbound from its storage resources.

    If storage resources are configured to be deleted upon clearing, their deletion
    occurs before transitioning to `CLEARED`. Otherwise, no actual work is required,
    and the transition happens immediately.

    If storage is not deleted during clearing, the responsibility to manage or delete
    those resources lies with the user.
    """

    @staticmethod
    def from_str(value):
        for member in StorageStatus:
            if member.name.lower() == value.lower():
                return member
        raise ValueError(f"Unknown value '{value}' for enum {StorageStatus.__name__}")

    def __eq__(self, other):
        return self.value == other.value


class FaultToleranceModel(Enum):
    """
    The fault tolerance model.
    """

    AtLeastOnce = 1
    """
    Each record is output at least once.  Crashes may duplicate output, but
    no input or output is dropped.
    """

    ExactlyOnce = 2
    """
    Each record is output exactly once.  Crashes do not drop or duplicate
    input or output.
    """

    def __str__(self) -> str:
        match self:
            case FaultToleranceModel.AtLeastOnce:
                return "at_least_once"
            case FaultToleranceModel.ExactlyOnce:
                return "exactly_once"

    @staticmethod
    def from_str(value):
        for member in FaultToleranceModel:
            if str(member) == value.lower():
                return member

        raise ValueError(
            f"Unknown value '{value}' for enum {FaultToleranceModel.__name__}"
        )


class PipelineFieldSelector(Enum):
    ALL = "all"
    """Select all fields of a pipeline."""

    STATUS = "status"
    """Select only the fields required to know the status of a pipeline."""


class BootstrapPolicy(Enum):
    AWAIT_APPROVAL = "await_approval"
    ALLOW = "allow"
    REJECT = "reject"



import logging
import os


def client_version() -> str:
    from importlib.metadata import version, PackageNotFoundError

    try:
        version = version("feldera")
    except PackageNotFoundError:
        version = "unknown"

    return version


def requests_verify_from_env() -> str | bool:
    env_feldera_tls_insecure = os.environ.get("FELDERA_TLS_INSECURE")
    FELDERA_HTTPS_TLS_CERT = os.environ.get("FELDERA_HTTPS_TLS_CERT")

    if env_feldera_tls_insecure is not None and FELDERA_HTTPS_TLS_CERT is not None:
        logging.warning(
            "environment variables FELDERA_HTTPS_TLS_CERT and "
            + "FELDERA_TLS_INSECURE both are set."
            + "\nFELDERA_HTTPS_TLS_CERT takes priority."
        )

    if env_feldera_tls_insecure is None:
        feldera_tls_insecure = False
    else:
        feldera_tls_insecure = env_feldera_tls_insecure.strip().lower() in (
            "1",
            "true",
            "yes",
        )

    requests_verify = not feldera_tls_insecure
    if FELDERA_HTTPS_TLS_CERT is not None:
        requests_verify = FELDERA_HTTPS_TLS_CERT

    return requests_verify



from typing import Optional
import os
from feldera.rest._helpers import requests_verify_from_env
import logging


class Config:
    """
    :class:`.FelderaClient`'s credentials and configuration parameters
    """

    def __init__(
        self,
        url: Optional[str] = None,
        api_key: Optional[str] = None,
        version: Optional[str] = None,
        timeout: Optional[float] = None,
        connection_timeout: Optional[float] = None,
        requests_verify: Optional[bool | str] = None,
    ) -> None:
        """
        :param url: The url to the Feldera API (ex: https://try.feldera.com)
        :param api_key: The optional API key to access Feldera
        :param version: The version of the API to use
        :param timeout: The timeout for the HTTP requests
        :param connection_timeout: The connection timeout for the HTTP requests
        :param requests_verify: The `verify` parameter passed to the requests
            library. `True` by default. Can also be set using environment
            variables `FELDERA_TLS_INSECURE` to disable TLS and
            `FELDERA_HTTPS_TLS_CERT` to set the certificate path. The latter
            takes priority.
        """

        BASE_URL = (
            url
            or os.environ.get("FELDERA_HOST")
            or os.environ.get("FELDERA_BASE_URL")
            or "http://localhost:8080"
        )
        self.url: str = BASE_URL
        self.api_key: Optional[str] = os.environ.get("FELDERA_API_KEY", api_key)
        self.version: str = version or "v0"
        self.timeout: Optional[float] = timeout
        self.connection_timeout: Optional[float] = connection_timeout
        env_verify = requests_verify_from_env()
        self.requests_verify: bool | str = (
            requests_verify if requests_verify is not None else env_verify
        )

        if self.requests_verify is False:
            logging.warning("TLS verification is disabled.")



import pandas as pd

from typing import Optional

from feldera import FelderaClient
from feldera._callback_runner import CallbackRunner


class OutputHandler:
    def __init__(
        self,
        client: FelderaClient,
        pipeline_name: str,
        view_name: str,
    ):
        """
        Initializes the output handler, but doesn't start it.
        To start the output handler, call the `.OutputHandler.start` method.
        """

        self.client: FelderaClient = client
        self.pipeline_name: str = pipeline_name
        self.view_name: str = view_name
        self.buffer: list[pd.DataFrame] = []
        self.exception: Optional[BaseException] = None

        # the callback that is passed to the `CallbackRunner`
        def callback(df: pd.DataFrame, _: int):
            if not df.empty:
                self.buffer.append(df)

        def exception_callback(exception: BaseException):
            self.exception = exception

        # sets up the callback runner
        self.handler = CallbackRunner(
            self.client,
            self.pipeline_name,
            self.view_name,
            callback,
            exception_callback,
        )

    def start(self):
        """
        Starts the output handler in a separate thread
        """

        self.handler.start()

    def to_pandas(self, clear_buffer: bool = True):
        """
        Returns the output of the pipeline as a pandas DataFrame

        :param clear_buffer: Whether to clear the buffer after getting the output.
        """

        if self.exception is not None:
            raise self.exception
        if len(self.buffer) == 0:
            return pd.DataFrame()
        res = pd.concat(self.buffer, ignore_index=True)
        if clear_buffer:
            self.buffer.clear()

        return res

    def to_dict(self, clear_buffer: bool = True):
        """
        Returns the output of the pipeline as a list of python dictionaries

        :param clear_buffer: Whether to clear the buffer after getting the output.
        """

        return self.to_pandas(clear_buffer).to_dict(orient="records")
